import sys
import inspect


def normalize(mat):
    """ Normalize columns so that each of them sum up to 1 """
    norms = mat.sum(0)
    norms[norms == 0] = 1
    mat /= norms
    return mat


def call_ignore_extra_args(func_base):
    def wrapper(self, obj, kwargs={}):
        func = getattr(self, func_base.__name__)
        internal_argnames = set(inspect.getargspec(func).args) - {'self'}
        internal_kwargs = {name: (getattr(obj, name) if hasattr(obj, name) else kwargs[name])
                           for name in internal_argnames}
        return func(**internal_kwargs)
    return wrapper


def public(f):
    """"Use a decorator to avoid retyping function/class names.

    * Based on an idea by Duncan Booth:
    http://groups.google.com/group/comp.lang.python/msg/11cbb03e09611b8a
    * Improved via a suggestion by Dave Angel:
    http://groups.google.com/group/comp.lang.python/msg/3d400fb22d8a42e1
    """
    all = sys.modules[f.__module__].__dict__.setdefault('__all__', [])
    if f.__name__ not in all:  # Prevent duplicates if run from an IDE.
        all.append(f.__name__)
    return f



from time import time
import numpy as np
import ctypes as c
import scipy.sparse
import numexpr as ne
import warnings
from ipy_progressbar import ProgressBar
from .regularizer import RegularizerWithCoefficient
from .quantity import QuantityBase
from .stop_condition import StopConditionBase
from ..utils import normalize, public

# try:
import pyximport
pyximport.install()
import cython_support
# except ImportError:
#     warnings.warn('Could not import MKL-based code, computations with sparse matrices are disabled.')


try:
    mkl = c.cdll.LoadLibrary("libmkl_rt.so")
except OSError:
    warnings.warn('Could not import MKL, using dot() from numpy.')
    dot = np.dot
else:
    def dot(A, B, C=None):
        assert A.shape[1] == B.shape[0]
        if C is None:
            C = np.empty((A.shape[0], B.shape[1]), dtype=np.float32)

        Order = 101  # 101 for row-major, 102 for column major data structures
        alpha = 1
        beta = 0

        m = A.shape[0]
        k = A.shape[1]
        n = B.shape[1]
        ldc = n

        if A.flags.c_contiguous:
            TransA = 111 # 111 for no transpose, 112 for transpose, and 113 for conjugate transpose
            lda = k
        elif A.flags.f_contiguous:
            TransA = 112
            lda = m

        if B.flags.c_contiguous:
            TransB = 111 # 111 for no transpose, 112 for transpose, and 113 for conjugate transpose
            ldb = n
        elif B.flags.f_contiguous:
            TransB = 112
            ldb = k

        mkl.cblas_sgemm( c.c_int(Order), c.c_int(TransA), c.c_int(TransB),
                         c.c_int(m), c.c_int(n), c.c_int(k),
                         c.c_float(alpha),
                         A.ctypes.data_as(c.POINTER(c.c_float)), c.c_int(lda),
                         B.ctypes.data_as(c.POINTER(c.c_float)), c.c_int(ldb),
                         c.c_float(beta),
                         C.ctypes.data_as(c.POINTER(c.c_float)), c.c_int(ldc) )
        return C


# def dot(A, B, C=None):
#     assert A.shape[1] == B.shape[0]
#     if C is None:
#         C = np.zeros((A.shape[0], B.shape[1]), dtype=np.float32)
#     for i in range(A.shape[0]):
#         for j in range(B.shape[1]):
#             for k in range(A.shape[1]):
#                 C[i, j] += A[i, k] * B[k, j]
#     return C


@public
class PlsaEmRational(object):

    """
    ### Non-matrix form
    * zero all $n_{wt}, n_{dt}, n_{t}$
    * for all $d, w$:
      * $Z = \sum_t \phi_{wt} \theta_{td}$
      * for all $t$:
        * increase $n_{wt}, n_{dt}, n_{t}$ by $\delta = n_{dw} \phi_{wt} \theta_{td} / Z$
    * $\phi_{wt} \propto \left( n_{wt} + \phi_{wt} \frac{\partial R}{\partial \phi_{wt}} \right)_+$
    * $\theta_{td} \propto \left( n_{dt} + \theta_{td} \frac{\partial R}{\partial \theta_{td}} \right)_+$

    ### Matrix form
    * $P_{wd} = \Phi_{wt} \Theta_{td}$
    * $P'_{wd} = N_{wd} / P_{wd}$
    * $N_{wt} = \Phi_{wt} \cdot P'_{wd} \Theta_{td}^T$
    * $N_{td} = \Theta_{td} \cdot \Phi_{wt}^T P'_{wd}$
    * $\Phi_{wt} \propto \left( N_{wt} + \Phi_{wt} \cdot \frac{\partial R}{\partial \Phi_{wt}} \right)_+ = \Phi_{wt} \cdot \left( P'_{wd} \Theta_{td}^T + \frac{\partial R}{\partial \Phi_{wt}} \right)_+$
    * $\Theta_{td} \propto \left( N_{td} + \Theta_{td} \cdot \frac{\partial R}{\partial \Theta_{td}} \right)_+ = \Theta_{td} \cdot \left( \Phi_{wt}^T P'_{wd} + \frac{\partial R}{\partial \Theta_{td}} \right)_+$
    """

    def __init__(self, nwd, T_init, modifiers):
        self.nwd = nwd
        self.W, self.D = self.nwd.shape
        self.T_init = T_init
        self.issparse = scipy.sparse.issparse(self.nwd)

        def type_checker(t):
            return lambda obj: isinstance(obj, t)
        self.regularizers = filter(type_checker(RegularizerWithCoefficient), modifiers)
        self.quantities = filter(type_checker(QuantityBase), modifiers)
        self.stop_conditions = filter(type_checker(StopConditionBase), modifiers)

        self.progress = []

    def generate_initial(self):
        self.phi = np.random.random((self.W, self.T_init)).astype(np.float32)
        normalize(self.phi)

        self.theta = np.random.random((self.T_init, self.D)).astype(np.float32)
        self.theta = np.asfortranarray(self.theta)
        normalize(self.theta)

        self.nd = np.squeeze(np.array(self.nwd.sum(0)))
        self.nw = np.squeeze(np.array(self.nwd.sum(1)))
        self.n = self.nwd.sum()

        # preallocate arrays once
        if not self.issparse:
            self.pwd = np.empty_like(self.nwd)
            self.npwd = np.empty_like(self.nwd)
        else:
            self.pwd = None
            self.npwd = None

    def calc_multipliers(self):
        if not self.issparse:
            ne.evaluate('where(nwd * pwd > 0, nwd / pwd, 0)', out=self.npwd, local_dict={'nwd': self.nwd, 'pwd': self.pwd})
            self.phi_sized = dot(self.npwd, self.theta.T)
            self.theta_sized = dot(self.phi.T, self.npwd)
        else:
            self.phi_sized = np.empty_like(self.phi)
            self.theta_sized = np.empty_like(self.theta)
            cython_support.calc_nwt(
                nwd=self.nwd,
                phi=self.phi,
                theta=self.theta,
                nwt_out=self.phi_sized)
            cython_support.calc_ntd(
                ndw=self.nwd.T,
                phi=self.phi,
                theta=self.theta,
                ntd_out=self.theta_sized)

    def iteration(self):
        if not self.issparse and self.itnum == 0:
            dot(self.phi, self.theta, self.pwd)

        self.calc_multipliers()

        dr_dphi = sum(reg.dr_dphi(self) for reg in self.regularizers)
        self.phi_new = self.phi * np.clip(self.phi_sized + dr_dphi, 0, float('inf'))

        dr_dtheta = 1.0 * self.n / self.D * sum(reg.dr_dtheta(self) for reg in self.regularizers)
        self.theta_new = self.theta * np.clip(self.theta_sized + dr_dtheta, 0, float('inf'))

        nonzero_t = ~np.all(self.theta_new == 0, axis=1)
        if nonzero_t.sum() > 0 and nonzero_t.sum() < self.theta.shape[0] / 1.1:
            self.phi_new = np.ascontiguousarray(self.phi_new[:, nonzero_t])
            self.theta_new = np.asfortranarray(self.theta_new[nonzero_t, :])

        self.phi = normalize(self.phi_new)
        self.theta = np.asfortranarray(normalize(self.theta_new))

        np.putmask(self.phi, self.phi < 1e-18, 0)
        np.putmask(self.theta, self.theta < 1e-18, 0)

        if not self.issparse:
            dot(self.phi, self.theta, self.pwd)

    def iterate(self, maxiter, prefunc=lambda s: s, quiet=False):
        self.maxiter = maxiter
        self.generate_initial()
        prefunc(self)
        try:
            pb = ProgressBar(range(maxiter), quiet=quiet >= 2, title='PLSA EM', key='plsa_em')

            for self.itnum in pb:
                start_time = time()
                self.iteration()
                end_time = time()

                progress_items = [
                    ('iteration', self.itnum),
                    ('time', end_time - start_time),
                ] + [(k, v) for q in self.quantities for k, v in q.items(self)]
                self.progress.append(progress_items)

                pb.set_extra_text('; '.join(['%s: %s' % (k.capitalize(), v) for k, v in progress_items[2:] if np.isscalar(v)]))
                if not quiet:
                    print '; '.join(['%s: %s' % (k.capitalize(), v) for k, v in progress_items])

                if any(sc.is_stop(self, dict(progress_items)) for sc in self.stop_conditions):
                    break
        except KeyboardInterrupt:
            print '<Interrupted>'



import numpy as np
from ..plsa import QuantityBase
from ..utils import public


@public
class TopicsLeft(QuantityBase):

    def _items(self, theta):
        yield ('topics_left', np.count_nonzero(theta.sum(1)))



from ..utils import call_ignore_extra_args, public


@public
class QuantityBase(object):

    def __init__(self):
        pass

    def _items(self):
        return []

    items = call_ignore_extra_args(_items)



from ..utils import call_ignore_extra_args, public
from .quantity import QuantityBase


@public
class RegularizerBase(object):

    def __init__(self):
        pass

    def _dr_dphi(self):
        return 0

    dr_dphi = call_ignore_extra_args(_dr_dphi)

    def _dr_dtheta(self):
        return 0

    dr_dtheta = call_ignore_extra_args(_dr_dtheta)


@public
class RegularizerCoefficientBase(object):

    def __init__(self):
        pass

    def _coefficient(self):
        return 0

    coefficient = call_ignore_extra_args(_coefficient)

    def __mul__(self, other):
        if isinstance(other, RegularizerBase):
            return RegularizerWithCoefficient(other, self)
        raise NotImplemented


@public
class RegularizerWithCoefficient(QuantityBase):

    def __init__(self, regularizer, coefficient):
        self.regularizer = regularizer
        self.coefficient = coefficient

    def dr_dphi(self, plsa):
        base_val = self.regularizer.dr_dphi(plsa)
        coeff = self.coefficient.coefficient(plsa)
        return coeff * base_val

    def dr_dtheta(self, plsa):
        base_val = self.regularizer.dr_dtheta(plsa)
        coeff = self.coefficient.coefficient(plsa)
        return coeff * base_val

    def items(self, plsa):
        coeff = self.coefficient.coefficient(plsa)
        yield ('coeff_%s' % self.regularizer.__class__.__name__, coeff)



from functools import partial
from ..plsa import StopConditionBase
from ..utils import public


@public
class ValueBounds(StopConditionBase):

    def __init__(self, value_name, lo=float('-inf'), hi=float('inf')):
        exec "self._is_stop = (lambda lo=lo, hi=hi: lambda %(value_name)s: not (lo <= %(value_name)s <= hi))()" % locals()



import math
import numpy as np
import numexpr as ne
import warnings
from ..plsa import QuantityBase
from ..utils import public

# try:
import pyximport
pyximport.install()
import perplexity_cython
# except ImportError:
#     warnings.warn('Could not import MKL-based code, computations with sparse matrices are disabled.')


@public
class Perplexity(QuantityBase):

    def __init__(self, exact=False):
        self.exact = exact

    def _items(self, n, nwd, phi, theta, pwd, issparse):
        if not issparse:
            if self.exact:
                s = ne.evaluate('where(nwd == 0, 0, nwd * log(pwd))').sum()
            else:
                mat = ne.evaluate('nwd * (pwd_i * a + b)',
                                  local_dict={'nwd': nwd,
                                              'a': np.float32(8.2629582881927490e-8),
                                              'b': np.float32(-87.989971088),
                                              'pwd_i': pwd.view(np.int32)},
                                  casting='unsafe')
                s = np.einsum('ij -> ', mat)
        else:
            # pwd is None here
            s = perplexity_cython.perplexity_sparse(nwd, phi, theta)

        yield ('perplexity', math.exp(-1 / n * s))



import numpy as np
from ..plsa import RegularizerBase
from ..utils import public


@public
class PhiSparse(RegularizerBase):
    r"""
    ### $\Phi$ Sparsifying regularizer
    * $R = -\sum_{wt} \ln \phi_{wt}$
    * $\frac{\partial R}{\partial \Phi_{td}} = -\frac{1}{\Phi_{wt}}$
    """
    def _dr_dphi(self, phi):
        return -1/phi


#cython: boundscheck=False, wraparound=False, embedsignature=True, cdivision=True
import numpy as np
cimport numpy as np


def perplexity_internal_cython(int n, float[:, :] pwd, float[:, :] nwd):
    cdef int* pwd_ = <int*> &pwd[0, 0]
    cdef float* nwd_ = &nwd[0, 0]
    cdef float res = 0
    cdef int i, j
    for i in range(n):
        res += nwd_[i] * (pwd_[i] * <float>8.2629582881927490e-8 - <float>87.989971088)
    return res


def perplexity_sparse(nwd,
                      float[:, ::1] phi,
                      float[::1, :] theta):
    nwd = nwd.tocsr()
    cdef int[:] nwd_indptr = nwd.indptr
    cdef int[:] nwd_indices = nwd.indices
    cdef float[:] nwd_data = nwd.data

    cdef int W = phi.shape[0]
    cdef int T = phi.shape[1]
    cdef int D = theta.shape[1]

    cdef int w, i, d
    cdef int i_0, i_1

    cdef float pwd_val
    cdef float result = 0

    for w in range(W):
        i_0 = nwd_indptr[w]
        i_1 = nwd_indptr[w + 1]

        for i in range(i_0, i_1):
            d = nwd_indices[i]
            pwd_val = np.dot(phi[w, :], theta[:, d])
            result += nwd_data[i] * ((<int*>&pwd_val)[0] * <float>8.2629582881927490e-8 - <float>87.989971088)

    return result



from ...plsa import RegularizerCoefficientBase
from ...utils import public


@public
class ZeroThenConstant(RegularizerCoefficientBase):

    def __init__(self, steps_zero, value):
    	self.steps_zero = steps_zero
        self.value = value

    def _coefficient(self, itnum):
        return self.value if itnum >= self.steps_zero else 0


@public
class Linear(RegularizerCoefficientBase):

    def __init__(self, max_value):
        self.max_value = max_value

    def _coefficient(self, itnum, maxiter):
        return self.max_value * 1.0 * itnum / maxiter



#cython: boundscheck=False, wraparound=False, embedsignature=True, cdivision=True
import numpy as np
cimport numpy as np


cdef saxpy(float a, float[:] x, float[:] y):
    for i in range(x.shape[0]):
        y[i] += a * x[i]


def calc_nwt(nwd,
             float[:, ::1] phi,
             float[::1, :] theta,
             float[:, ::1] nwt_out):
    nwd = nwd.tocsr()
    cdef int[:] nwd_indptr = nwd.indptr
    cdef int[:] nwd_indices = nwd.indices
    cdef float[:] nwd_data = nwd.data

    cdef int W = phi.shape[0]
    cdef int T = phi.shape[1]
    cdef int D = theta.shape[1]

    cdef int w, t, i, d
    cdef int i_0, i_1
    cdef float pwd_val

    for w in range(W):
        i_0 = nwd_indptr[w]
        i_1 = nwd_indptr[w + 1]

        for t in range(T):
            nwt_out[w, t] = 0

        for i in range(i_0, i_1):
            d = nwd_indices[i]
            pwd_val = np.dot(phi[w, :], theta[:, d])
            if pwd_val == 0:
                continue
            saxpy(nwd_data[i] / pwd_val, theta[:, d], nwt_out[w, :])


def calc_ntd(ndw,
             float[:, ::1] phi,
             float[::1, :] theta,
             float[::1, :] ntd_out):
    ndw = ndw.tocsr()
    cdef int[:] ndw_indptr = ndw.indptr
    cdef int[:] ndw_indices = ndw.indices
    cdef float[:] ndw_data = ndw.data

    cdef int W = phi.shape[0]
    cdef int T = phi.shape[1]
    cdef int D = theta.shape[1]

    cdef int w, t, i, d
    cdef int i_0, i_1
    cdef float pwd_val

    for d in range(D):
        i_0 = ndw_indptr[d]
        i_1 = ndw_indptr[d + 1]

        for t in range(T):
            ntd_out[t, d] = 0

        for i in range(i_0, i_1):
            w = ndw_indices[i]
            pwd_val = np.dot(phi[w, :], theta[:, d])
            if pwd_val == 0:
                continue
            saxpy(ndw_data[i] / pwd_val, phi[w, :], ntd_out[:, d])



from ..utils import call_ignore_extra_args, public


class StopConditionBase(object):

    def __init__(self):
        pass

    def _is_stop(self):
        return []

    is_stop = call_ignore_extra_args(_is_stop)



import numpy as np
from ..plsa import RegularizerBase
from ..utils import public


@public
class ThetaLineSparse(RegularizerBase):
    r"""
    ### Line sparsifying regularizer
    * $R = \frac{1}{|T|} \sum_t \ln \frac{1/|T|}{\sum_d \theta_{td} n_d / n}$
    * $\frac{\partial R}{\partial \Theta_{td}} = -\frac{1}{|T|} \frac{n_d}{\sum_{d'} \theta_{td'} n_{d'}} = -\frac{1}{|T|} \frac{N_d}{(\Theta N)_t}$

    ### Another version
    * $R = \sum_t \left( \sum_d \theta_{td} n_d / n \right) \ln \frac{\sum_d \theta_{td} n_d / n}{1/|T|} \propto \sum_t \left( \sum_d \theta_{td} n_d \right) (\ln \sum_d \theta_{td} n_d - \ln n/|T|)$
    * $\frac{\partial R}{\partial \Theta_{td}} = n_d \left( 1 + \ln \sum_{d'} \theta_{td'} n_{d'} - \ln \frac{n}{T} \right) = N_d \left( 1 + \ln (\Theta N)_t - \ln \frac{n}{T} \right)$
    """
    def _dr_dtheta(self, T_init, nd, D, theta):
        return -1.0 * D / T_init * nd / np.dot(theta, nd.reshape((-1, 1)))


import numpy as np
from ..plsa import RegularizerBase
from ..utils import public


@public
class ThetaSparse(RegularizerBase):
    r"""
    ### $\Theta$ Sparsifying regularizer
    * $R = -\sum_{td} \ln \theta_{td}$
    * $\frac{\partial R}{\partial \Theta_{td}} = -\frac{1}{\Theta_{td}}$
    """
    def _dr_dtheta(self, theta):
        return -1/theta



from setuptools import setup


try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except (IOError, ImportError):
    long_description = ''

setup(name='py-artm',
      version='0.2.2',
      description='',
      long_description = long_description,
      classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Software Development :: Libraries :: Python Modules',
      ],
      keywords='',
      url='',
      author='Alexander Plavin',
      author_email='alexander@plav.in',
      license='MIT',
      packages=['py_artm', 'py_artm.plsa', 'py_artm.regularizers', 'py_artm.regularizers.coefficients', 'py_artm.quantities', 'py_artm.stop_conditions'],
      package_data={'py_artm': ['**/*.pyx', '**/*.pyxbld']},
      install_requires=[ 'numpy', 'scipy', 'numexpr', 'ipy-progressbar' ],
      zip_safe=False)



from ...plsa import RegularizerCoefficientBase
from ...utils import public


@public
class Constant(RegularizerCoefficientBase):

    def __init__(self, value):
        self.value = value

    def _coefficient(self):
        return self.value
